#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.signals as signals

from tests.signals import signal_dict, signal_bundle_dict


def test_create_signal_bundle(signal_bundle_dict):
    created_bundle = signals.create_signal_bundle(signal_bundle_dict)
    assert len(created_bundle.signals) == 1
    assert created_bundle.identifier == "dict identifier"
    assert created_bundle.version == "dict_version"
    assert created_bundle.signals[0].topic == "dict topic"
    assert created_bundle.signals[0].content == {"dict": "content", "hi": 1}


def test_create_signal(signal_dict):
    created_signal = signals.create_signal(signal_dict)
    assert created_signal.topic == "dict topic"
    assert created_signal.content == {"dict": "content", "hi": 1}



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.display import display_translator
from octobot_commons.display.display_translator import (
    DisplayTranslator,
    Element,
)

from octobot_commons.display import display_factory
from octobot_commons.display.display_factory import (
    display_translator_factory,
)

from octobot_commons.display import plot_settings
from octobot_commons.display.plot_settings import (
    PlotSettings,
)


__all__ = ["DisplayTranslator", "Element", "display_translator_factory", "PlotSettings"]



# pylint: disable=W0212
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses


@dataclasses.dataclass
class UpdatableDataclass:
    _updated_fields: list[str] = dataclasses.field(default_factory=list, kw_only=True)

    def update(self, other) -> None:
        """
        update self using another UpdatableDataclass
        :param other: the other UpdatableDataclass to update self from
        """
        for field_name in other._updated_fields:
            self_val = getattr(self, field_name)
            other_val = getattr(other, field_name)
            if isinstance(self_val, UpdatableDataclass) and other_val:
                self_val.update(other_val)
            elif (
                isinstance(self_val, list)
                and self_val
                and isinstance(self_val[0], UpdatableDataclass)
            ):
                updated_list = []
                for i, other_element in enumerate(other_val):
                    if i < len(self_val):
                        self_element = self_val[i]
                        self_element.update(other_element)
                        updated_list.append(self_element)
                    else:
                        updated_list.append(other_element)
                setattr(self, field_name, updated_list)
            elif _should_be_changed(self_val, other_val):
                setattr(self, field_name, other_val)

    def get_update(self, other):
        """
        Creates a new instance of self.__class__ which fields will be set only if they changed between self and other
        Requires a default constructor
        :param other: the other UpdatableDataclass create the update from
        :return: the UpdatableDataclass update containing differences between self and other
        (unset values in other are ignored)
        """
        update_content = self.__class__()
        for field in dataclasses.fields(self):
            field_name = field.name
            self_val = getattr(self, field_name)
            other_val = getattr(other, field_name)
            if isinstance(self_val, UpdatableDataclass) and other_val:
                update = self_val.get_update(other_val)
                setattr(update_content, field_name, update)
                if update._updated_fields:
                    update_content._updated_fields.append(field_name)
            elif (
                isinstance(self_val, list)
                and self_val
                and isinstance(self_val[0], UpdatableDataclass)
            ):
                update_list = []
                has_updates = False
                if self_val != other_val:
                    for i, other_element in enumerate(other_val):
                        if i < len(self_val):
                            self_element = self_val[i]
                            update = self_element.get_update(other_element)
                            update_list.append(update)
                            has_updates = has_updates or bool(update._updated_fields)
                        else:
                            update_list.append(other_element)
                            has_updates = True
                setattr(update_content, field_name, update_list)
                if has_updates or len(other_val) != len(self_val):
                    update_content._updated_fields.append(field_name)
            elif _should_be_changed(self_val, other_val):
                update_content._updated_fields.append(field_name)
                setattr(update_content, field_name, other_val)
        return update_content


def _should_be_changed(current_value, new_value):
    return (
        not current_value and new_value != current_value
    ) or new_value != current_value



# pylint: disable=C0301,R0902,R0913,C0415,R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.errors as errors
import octobot_commons.databases.cache_manager as cache_manager
import octobot_commons.databases.implementations as implementations
import octobot_commons.databases.document_database_adaptors as document_database_adaptors


class CacheClient:
    def __init__(
        self,
        tentacle,
        exchange_name,
        symbol,
        time_frame,
        tentacles_setup_config,
        flush_cache_when_necessary,
        config_name=None,
    ):
        self._flush_cache_when_necessary = flush_cache_when_necessary
        self.cache_manager = cache_manager.CacheManager(
            database_adaptor=document_database_adaptors.TinyDBAdaptor
        )
        self.config_name = config_name or self.cache_manager.DEFAULT_CONFIG_IDENTIFIER
        self.tentacle = tentacle
        self.exchange_name = exchange_name
        self.symbol = symbol
        self.time_frame = time_frame
        self.tentacles_setup_config = tentacles_setup_config
        try:
            import octobot_tentacles_manager.models as tentacles_manager_models

            self.tentacles_requirements = (
                tentacles_manager_models.TentacleRequirementsTree(
                    self.tentacle, self.config_name
                )
            )
        except ImportError as err:
            raise ImportError(
                "OctoBot-Tentacles-Manager is required to use cache clients"
            ) from err

    def has_cache(self, pair, time_frame, tentacle_name=None, config_name=None):
        """
        Returns True when a cache as specified in arguments is currently open
        :param pair: pair/symbol to build a cache path from
        :param pair: time_frame to build a cache path from
        :param tentacle_name: tentacle to build a cache path from
        :param config_name: name of the configuration
        """
        return self.cache_manager.has_cache(
            tentacle_name or self.tentacle.get_name(),
            self.exchange_name,
            pair,
            time_frame,
            config_name or self.config_name,
        )

    def get_cache_path(self, tentacle, config_name=None):
        """
        Returns the path to the cache associated to the given tentacle
        :param tentacle: tentacle to build a cache path from
        :param config_name: name of the configuration
        """
        config_name = config_name or self.config_name
        return self.cache_manager.get_cache_or_build_path(
            tentacle,
            self.exchange_name,
            self.symbol,
            self.time_frame,
            tentacle.get_name(),
            config_name,
            self.tentacles_setup_config,
            self.tentacles_requirements,
        )

    def get_cache(
        self,
        tentacle_name=None,
        cache_type=implementations.CacheTimestampDatabase,
        config_name=None,
    ):
        """
        Returns the cache associated to the given tentacle_name
        :param tentacle_name: name of the tentacle to get cache from
        :param cache_type: type of the cache
        :param config_name: name of the configuration
        """
        tentacle = self.tentacle if tentacle_name is None else None
        tentacle_name = tentacle_name or self.tentacle.get_name()
        config_name = config_name or self.config_name
        cache, just_created = self.cache_manager.get_cache(
            tentacle,
            tentacle_name,
            self.exchange_name,
            self.symbol,
            self.time_frame,
            config_name,
            self.tentacles_setup_config,
            self.tentacles_requirements,
            cache_type=cache_type,
        )
        if just_created and cache_type is implementations.CacheTimestampDatabase:
            if tentacle is None:
                metadata = self.cache_manager.get_cache_previous_db_metadata(
                    tentacle_name,
                    self.exchange_name,
                    self.symbol,
                    self.time_frame,
                    config_name,
                )
            else:
                metadata = {
                    enums.CacheDatabaseColumns.TRIGGERED_AFTER_CANDLES_CLOSE.value: tentacle.is_triggered_after_candle_close
                }
            if metadata is None:
                raise RuntimeError(
                    "Missing db metadata. Please provide the tentacle parameter to this method"
                )
            cache.add_metadata(metadata)
        return cache

    async def get_cached_value(
        self,
        value_key: str = enums.CacheDatabaseColumns.VALUE.value,
        cache_key=None,
        tentacle_name=None,
        config_name=None,
    ) -> tuple:
        """
        Get a value for the current cache
        :param value_key: identifier of the value
        :param cache_key: timestamp to use in order to look for a value
        :param tentacle_name: name of the tentacle to get cache from
        :param config_name: name of the tentacle configuration as used in nested tentacle calls
        :return: the cached value and a boolean (True if cached value is missing from cache)
        """
        try:
            return (
                await self.get_cache(
                    tentacle_name=tentacle_name, config_name=config_name
                ).get(cache_key, name=value_key),
                False,
            )
        except errors.NoCacheValue:
            return None, True

    async def set_cached_value(
        self,
        value,
        value_key: str = enums.CacheDatabaseColumns.VALUE.value,
        cache_key=None,
        flush_if_necessary=False,
        tentacle_name=None,
        config_name=None,
        **kwargs,
    ):
        """
        Set a value into the current cache
        :param value: value to set
        :param value_key: identifier of the value
        :param cache_key: timestamp to associate the value to
        :param flush_if_necessary: flush the cache after set (write into database)
        :param tentacle_name: name of the tentacle to get cache from
        :param config_name: name of the tentacle configuration as used in nested tentacle calls
        :param kwargs: other related value_key / value couples to set at this timestamp. Use for plotted data
        :return: None
        """
        cache = None
        try:
            cache = self.get_cache(tentacle_name=tentacle_name, config_name=config_name)
            await cache.set(cache_key, value, name=value_key)
            if kwargs:
                for key, val in kwargs.items():
                    await cache.set(
                        cache_key,
                        val,
                        name=f"{value_key}{constants.CACHE_RELATED_DATA_SEPARATOR}{key}",
                    )
        finally:
            if flush_if_necessary and self._flush_cache_when_necessary and cache:
                await cache.flush()

    async def set_cached_values(
        self,
        values,
        value_key,
        cache_keys,
        flush_if_necessary=False,
        tentacle_name=None,
        config_name=None,
        additional_values_by_key=None,
    ):
        """
        Set a value into the current cache
        :param values: values to set
        :param value_key: identifier of the value
        :param cache_keys: timestamps to associate the values to
        :param flush_if_necessary: flush the cache after set (write into database)
        :param tentacle_name: name of the tentacle to get cache from
        :param config_name: name of the tentacle configuration as used in nested tentacle calls
        :param additional_values_by_key: other values to set in a dict of cache_keys
        :return: None
        """
        cache = None
        try:
            cache = self.get_cache(tentacle_name=tentacle_name, config_name=config_name)
            await cache.set_values(
                cache_keys,
                values,
                name=value_key,
                additional_values_by_key=additional_values_by_key,
            )
        finally:
            if flush_if_necessary and self._flush_cache_when_necessary and cache:
                await cache.flush()

    def ensure_no_missing_cached_value(self, is_missing):
        """
        Raises NoCacheValue when is_missing is True
        :param is_missing: True when a value is missing
        """
        if is_missing:
            raise errors.NoCacheValue(
                f"No cache value with cache key: {enums.CacheDatabaseColumns.VALUE.value}. "
                f"Impossible process {constants.DO_NOT_OVERRIDE_CACHE} return value."
            )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class Singleton:
    """
    From https://stackoverflow.com/questions/51245056/singleton-is-not-working-in-cython
    """

    _instances = {}

    @classmethod
    def instance(cls, *args, **kwargs):
        """
        Create the instance if not already created
        Return the class instance
        :param args: the constructor arguments
        :param kwargs: the constructor optional arguments
        :return: the class only instance
        """
        if cls not in cls._instances:
            cls._instances[cls] = cls(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def get_instance_if_exists(cls):
        """
        Return the instance if it exist
        Return the class instance if it exist
        :return: the class only instance if it exist otherwise None
        """
        try:
            return cls._instances[cls]
        except KeyError:
            return None



# pylint: disable=W0703
#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import threading
import csv
import gc

import octobot_commons.constants as commons_constants
import octobot_commons.singleton as singleton
import octobot_commons.timestamp_util as timestamp_util
import octobot_commons.logging as logging
import octobot_commons.async_job as async_job
import octobot_commons.os_util as os_util


class SystemResourcesWatcher(singleton.Singleton):
    DEFAULT_WATCHER_INTERVAL = (
        commons_constants.RESOURCES_WATCHER_MINUTES_INTERVAL
        * commons_constants.MINUTE_TO_SECONDS
    )
    CPU_WATCHING_SECONDS = 2

    def __init__(self, dump_resources, output_file):
        super().__init__()
        self.watcher_job = None
        self.watcher_interval = self.DEFAULT_WATCHER_INTERVAL
        self.logger = logging.get_logger(self.__class__.__name__)
        self.dump_resources = dump_resources
        self.output_file = output_file
        self.initialized_output = False

    def _exec_log_used_resources(self):
        try:
            # trigger garbage collector to get a fresh memory picture
            gc.collect()
            # warning: blocking to monitor CPU usage, to be used in a thread
            cpu, percent_ram, ram, process_ram = os_util.get_cpu_and_ram_usage(
                self.CPU_WATCHING_SECONDS
            )
            self.logger.debug(
                f"Used system resources: {cpu}% CPU, {round(ram, 3)} GB in RAM ({percent_ram}% of total "
                f"including {process_ram} GB from this process). "
            )
            if self.dump_resources:
                self._dump_resources(cpu, percent_ram, ram, process_ram)
        except Exception as err:
            self.logger.exception(err, False)
            self.logger.debug(f"Error when checking system resources: {err}")

    async def _log_used_resources(self):
        threading.Thread(
            target=self._exec_log_used_resources,
            daemon=True,
            name=f"{self.__class__.__name__}-_exec_log_used_resources",
        ).start()

    def _dump_resources(self, cpu, percent_ram, ram, process_ram):
        reset_file = not self.initialized_output
        self.initialized_output = True
        mode = "w" if reset_file else "a"
        row = (
            str(element).replace(".", ",")
            for element in (
                timestamp_util.get_now_time(),
                process_ram,
                cpu,
                percent_ram,
                ram,
            )
        )
        with open(self.output_file, mode, newline="") as csv_file:
            writer = csv.writer(csv_file, delimiter=";")
            if reset_file:
                writer.writerow(
                    [
                        "TIME",
                        "PROCESS USED RAM",
                        "% USED CPU",
                        "% USED RAM",
                        "TOTAL USED RAM",
                    ]
                )
            writer.writerow(row)

    async def start(self):
        """
        Synch the clock and start the clock synchronization loop if possible on this system
        """
        self.logger.debug("Starting system resources watcher")
        self.watcher_job = async_job.AsyncJob(
            self._log_used_resources,
            execution_interval_delay=self.watcher_interval,
        )
        await self.watcher_job.run()

    def stop(self):
        """
        Stop the synchronization loop
        """
        if self.watcher_job is not None and not self.watcher_job.is_stopped():
            self.logger.debug("Stopping system resources watcher")
            self.watcher_job.stop()


async def start_system_resources_watcher(dump_resources, output_file):
    """
    Start the resources watcher loop
    """
    await SystemResourcesWatcher.instance(dump_resources, output_file).start()


async def stop_system_resources_watcher():
    """
    Stop the watcher loop
    """
    return SystemResourcesWatcher.instance().stop()
